import functools
import logging
import re
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Tuple, Union

from RPA.core.locators import LocatorsDatabase, WindowsLocator
from RPA.core.vendor.deco import keyword as method
from RPA.core.windows.context import (
    ElementNotFound,
    WindowControlError,
    WindowsContext,
    with_timeout,
)
from RPA.core.windows.helpers import IS_WINDOWS

if IS_WINDOWS:
    import uiautomation as auto
    from uiautomation.uiautomation import Control


Locator = Union["WindowsElement", str]
SearchType = Dict[str, Union[str, int, List[int]]]


@dataclass
class WindowsElement:
    """Represent Control as dataclass"""

    item: "Control"
    locator: Optional[Locator] = None
    name: str = ""
    automation_id: str = ""
    control_type: str = ""
    class_name: str = ""
    left: int = -1
    right: int = -1
    top: int = -1
    bottom: int = -1
    width: int = -1
    height: int = -1
    xcenter: int = -1
    ycenter: int = -1

    def __init__(self, item: "Control", locator: Optional[Locator]):
        self.item: "Control" = item
        self.locator: Optional[Locator] = locator
        self.name = item.Name
        self.automation_id = item.AutomationId
        self.control_type = item.ControlTypeName
        self.class_name = item.ClassName
        # If there's no rectangle, then all coords are defaulting to -1.
        rect = item.BoundingRectangle
        if rect:
            self.left = rect.left
            self.right = rect.right
            self.top = rect.top
            self.bottom = rect.bottom
            self.width = rect.width()
            self.height = rect.height()
            self.xcenter = rect.xcenter()
            self.ycenter = rect.ycenter()

        self._sibling_element_compare = {
            # <locator_strategy>: <element_attribute>
            "id": "automation_id",
            "automationid": "automation_id",
            re.compile(r"(?<!sub)name:"): "name",
            "subname": self._cmp_subname,
            "regex": self._cmp_regex,
            "class": "class_name",
            "control": "control_type",
            "type": "control_type",
        }

    @staticmethod
    def _get_locator_value(locator: str, strategy: str) -> str:
        # pylint: disable=not-an-iterable
        for loc in MatchObject.parse_locator(locator).locators:
            if loc[0] == strategy:
                return loc[1]

        raise ValueError(
            f"couldn't find {strategy!r} in parsed final sub-locator: {locator}"
        )

    @classmethod
    def _cmp_subname(cls, win_elem: "WindowsElement", *, locator: str) -> bool:
        subname = cls._get_locator_value(locator, "SubName")
        return subname in win_elem.name

    @classmethod
    def _cmp_regex(cls, win_elem: "WindowsElement", *, locator: str) -> bool:
        pattern = cls._get_locator_value(locator, "RegexName")
        return bool(re.match(pattern, win_elem.name))

    @staticmethod
    def norm_locator(locator: Optional[Locator]) -> Optional[str]:
        while locator:
            if isinstance(locator, WindowsElement):
                locator = locator.locator
            else:  # finally, reached a string locator
                break
        return locator

    def is_sibling(self, win_elem: "WindowsElement") -> bool:
        """Returns `True` if the provided window element is a sibling."""
        locator: Optional[str] = self.norm_locator(win_elem)
        if not locator:
            return True  # nothing to check here, can be considered sibling

        last_locator_part = locator.split(MatchObject.TREE_SEP)[-1]
        cmp_attrs = []
        for strategy, attr_or_func in self._sibling_element_compare.items():
            if isinstance(strategy, str):
                strategy_regex = re.compile(rf"{strategy}:")
            else:
                strategy_regex = strategy
            if strategy_regex.search(last_locator_part):
                cmp_attrs.append(attr_or_func)
        # Name comparison is assumed by default if no strategies are found at all.
        cmp_attrs = cmp_attrs or ["name"]
        for attr_or_func in cmp_attrs:
            if isinstance(attr_or_func, str):
                status = getattr(self, attr_or_func) == getattr(win_elem, attr_or_func)
            else:
                status = attr_or_func(  # pylint: disable=not-callable
                    win_elem, locator=last_locator_part
                )
            if not status:
                return False

        return True


@dataclass
class MatchObject:
    """Represents all locator parts as object properties"""

    _WINDOWS_LOCATOR_STRATEGIES = {
        # RPA-strategy: UIA-strategy
        "automationid": "AutomationId",
        "id": "AutomationId",
        "class": "ClassName",
        "control": "ControlType",
        "depth": "searchDepth",
        "name": "Name",
        "regex": "RegexName",
        "subname": "SubName",
        "type": "ControlType",
        "index": "foundIndex",
        "offset": "offset",
        "desktop": "desktop",
        "process": "process",
        "handle": "handle",
        "executable": "executable",
        "path": "path",
    }
    TREE_SEP = " > "
    PATH_SEP = "|"  # path locator index separator
    QUOTE = '"'  # enclosing quote character
    _LOCATOR_REGEX = re.compile(rf"\S*{QUOTE}[^{QUOTE}]+{QUOTE}|\S+", re.IGNORECASE)
    _LOGGER = logging.getLogger(__name__)

    locators: List[Tuple] = field(default_factory=list)
    _classes: Set[str] = field(default_factory=set)
    max_level: int = 0

    @classmethod
    def parse_locator(cls, locator: str) -> "MatchObject":
        match_object = MatchObject()
        locator_tree = [loc.strip() for loc in locator.split(cls.TREE_SEP)]
        for level, branch in enumerate(locator_tree):
            default_values = []
            for part in cls._LOCATOR_REGEX.finditer(branch):
                match_object.handle_locator_part(
                    level, part.group().strip(), default_values
                )
            if default_values:
                match_object.add_locator("Name", " ".join(default_values), level=level)
        if not match_object.locators:
            match_object.add_locator("Name", locator)
        return match_object

    def handle_locator_part(
        self, level: int, part_text: str, default_values: List[str]
    ) -> None:
        if not part_text:
            return

        add_locator = functools.partial(self.add_locator, level=level)

        if part_text in ("and", "or", "desktop"):
            # NOTE(cmin764): Only "and" is supported at the moment. (match type is
            #  ignored and spaces are treated as "and"s by default)
            if part_text == "desktop":
                add_locator("desktop", "desktop")
            return

        try:
            strategy, value = part_text.split(":", 1)
        except ValueError:
            self._LOGGER.debug("No locator strategy found. (assuming 'name')")
            default_values.append(part_text)
            return

        control_strategy = self._WINDOWS_LOCATOR_STRATEGIES.get(strategy)
        if control_strategy:
            if default_values:
                add_locator("Name", " ".join(default_values))
                default_values.clear()
            add_locator(control_strategy, value)
        else:
            self._LOGGER.warning(
                "Invalid locator strategy %r! (assuming 'name')", strategy
            )
            default_values.append(part_text)

    def add_locator(self, control_strategy: str, value: str, level: int = 0) -> None:
        value = value.strip(f"{self.QUOTE} ")
        if not value:
            return

        self.max_level = max(self.max_level, level)

        if control_strategy in ("foundIndex", "searchDepth", "handle"):
            value = int(value)
        elif control_strategy == "ControlType":
            value = value if value.endswith("Control") else f"{value}Control"
        elif control_strategy == "ClassName":
            self._classes.add(value.lower())  # pylint: disable=no-member
        elif control_strategy == "path":
            value = [
                int(idx)
                for idx in value.strip(f" {self.PATH_SEP}").split(self.PATH_SEP)
            ]
        self.locators.append(  # pylint: disable=no-member
            (control_strategy, value, level)
        )

    @property
    def classes(self) -> List[str]:
        return list(self._classes)


class LocatorMethods(WindowsContext):
    """Keywords for finding Windows GUI elements"""

    def __init__(self, ctx, locators_path: Optional[str] = None):
        super().__init__(ctx)
        self._locators_path = locators_path

    @staticmethod
    def _get_desktop_control() -> "Control":
        root_control = auto.GetRootControl()
        new_control = Control.CreateControlFromControl(root_control)
        new_control.robocorp_click_offset = None
        return new_control

    @classmethod
    def get_desktop_element(cls, locator: Optional[Locator] = None) -> WindowsElement:
        desktop_control = cls._get_desktop_control()
        return WindowsElement(desktop_control, locator)

    @staticmethod
    def _get_control_from_params(
        search_params: SearchType, root_control: Optional["Control"] = None
    ) -> "Control":
        search_params = search_params.copy()  # to keep idempotent behaviour
        offset = search_params.pop("offset", None)
        control_type = search_params.pop("ControlType", "Control")
        ElementControl = getattr(root_control, control_type, Control)
        control = ElementControl(**search_params)
        new_control = Control.CreateControlFromControl(control)
        new_control.robocorp_click_offset = offset
        return new_control

    def _get_control_from_listed_windows(
        self, search_params: SearchType, *, param_type: str, win_type: str
    ) -> "Control":
        search_params = search_params.copy()  # to keep idempotent behaviour
        win_value = search_params.pop(param_type)
        window_list = self.ctx.list_windows()
        matches = [win for win in window_list if win[win_type] == win_value]
        if not matches:
            raise WindowControlError(
                f"Could not locate window with {param_type} {win_value!r}"
            )
        elif len(matches) > 1:
            raise WindowControlError(
                f"Found more than one window with {param_type} {win_value!r}"
            )
        self.logger.info("Found process with window title: %r", matches[0]["title"])
        search_params["Name"] = matches[0]["title"]
        return self._get_control_from_params(search_params)

    def _get_control_from_path(
        self, search_params: SearchType, root_control: "Control"
    ) -> "Control":
        # Follow a path in the tree of controls until reaching the final target.
        search_params = search_params.copy()  # to keep idempotent behaviour
        path = search_params["path"]
        current = root_control
        to_path = lambda idx: (  # noqa: E731
            MatchObject.PATH_SEP.join(str(pos) for pos in path[:idx])
        )

        for index, position in enumerate(path):
            children = current.GetChildren()
            if position > len(children):
                raise ElementNotFound(
                    f"Unable to retrieve child on position {position!r} under a parent"
                    f" with partial path {to_path(index)!r}"
                )

            current = children[position - 1]
            self.logger.debug(
                "On child position %d found control: %s", position, current
            )

        offset = search_params.get("offset")
        current.robocorp_click_offset = offset
        return current

    def _get_control_with_locator_part(
        self, locator: str, search_depth: int, root_control: "Control"
    ) -> "Control":
        # Prepare control search parameters.
        match_object = MatchObject.parse_locator(locator)
        self.logger.info("Locator %r produced matcher: %s", locator, match_object)
        search_params = {}
        for loc in match_object.locators:  # pylint: disable=not-an-iterable
            search_params[loc[0]] = loc[1]
        if "searchDepth" not in search_params:
            search_params["searchDepth"] = search_depth
        elif {"desktop", "path"} & set(search_params):
            self.logger.warning(
                "Depth strategy has no effect on 'desktop:' or 'path:' ones!"
            )

        # Obtain an element with the search parameters.
        if "desktop" in search_params:
            return self._get_desktop_control()

        if "executable" in search_params:
            return self._get_control_from_listed_windows(
                search_params, param_type="executable", win_type="name"
            )

        if "handle" in search_params:
            return self._get_control_from_listed_windows(
                search_params, param_type="handle", win_type="handle"
            )

        if "path" in search_params:
            return self._get_control_from_path(search_params, root_control)

        return self._get_control_from_params(search_params, root_control=root_control)

    def _load_by_alias(self, criteria: str) -> str:
        try:
            locator = LocatorsDatabase.load_by_name(criteria, self._locators_path)
            if isinstance(locator, WindowsLocator):
                return locator.value
        except ValueError:
            pass

        return criteria

    def _resolve_root(self, root_element: Optional[WindowsElement]) -> WindowsElement:
        # Explicit root element > set anchor > active window > Desktop.
        root = (
            self._window_or_none(root_element)
            or self.anchor
            or self.window
            or self.get_desktop_element()
        )
        self.logger.info("Resulted root element: %s", root)
        return root

    def _get_element_by_locator_string(
        self, locator: str, search_depth: int, root_element: Optional[WindowsElement]
    ) -> WindowsElement:
        root_control = self._resolve_root(root_element).item
        locator_parts = locator.split(MatchObject.TREE_SEP)
        assert locator_parts, "empty locator"

        try:
            for locator_part in locator_parts:
                self.logger.debug("Active root element: %r", root_control)
                control = self._get_control_with_locator_part(
                    locator_part, search_depth, root_control
                )
                root_control = control
        except LookupError as err:
            raise ElementNotFound(
                f"Element not found with locator {locator!r}"
            ) from err

        # If we get here, a `control` item was found.
        return WindowsElement(control, locator)

    @method
    @with_timeout
    def get_element(
        self,
        locator: Optional[Locator] = None,
        search_depth: int = 8,
        root_element: Optional[WindowsElement] = None,
        timeout: Optional[float] = None,  # pylint: disable=unused-argument
    ) -> WindowsElement:
        if isinstance(locator, str):
            locator = self._load_by_alias(locator)
        self.logger.info("Getting element with locator: %s", locator)
        if not locator:
            return self._resolve_root(root_element)
        elif isinstance(locator, str):
            element = self._get_element_by_locator_string(
                locator, search_depth, root_element
            )
        else:
            element = locator
        if self._window_or_none(element) is None:
            raise ElementNotFound(f"Unable to get element with {locator!r}")
        self.logger.info("Returning element: %s", element)
        return element



from typing import Optional, Union

from RPA.core.locators import LocatorsDatabase, Locator, TYPES


LocatorType = Union[str, Locator]


def _unquote(text):
    if text.startswith('"') and text.endswith('"'):
        text = text[1:-1]
    return text


def parse(locator: LocatorType, path: Optional[str] = None) -> Locator:
    """Parse locator string literal into a ``Locator`` instance.

    For example: "image:path/to/image.png" -> ImageLocator(path="path/to/image-png")
    """
    if isinstance(locator, Locator):
        return locator

    try:
        typename, _, value = str(locator).partition(":")
    except ValueError as err:
        raise ValueError(f"Invalid locator format: {locator}") from err

    if not value:
        typename, value = "alias", typename

    typename = typename.strip().lower()
    if typename == "alias":
        return LocatorsDatabase.load_by_name(_unquote(value), path)
    else:
        klass = TYPES.get(typename)
        if not klass:
            raise ValueError(f"Unknown locator type: {typename}")

        args = [_unquote(arg) for arg in value.split(",")]
        return klass(*args)



import base64
import io
import json
import logging
import os
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Union

from RPA.core.locators import BrowserLocator, ImageLocator, Locator


@contextmanager
def open_stream(path_or_stream, *args, encoding="utf-8", **kwargs) -> io.IOBase:
    """Wrapper for built-in open(), which allows using
    existing IO streams.
    """
    if not isinstance(path_or_stream, io.IOBase):
        path_or_stream = open(path_or_stream, *args, encoding=encoding, **kwargs)

    with path_or_stream:  # closes the descriptor even if it errors
        yield path_or_stream


def sanitize_name(name):
    """Sanitize locator name for use in filenames.
    Sanitized name should be unique in database.

    Examples:
        Google.Logo -> google-logo
        Cool Stuff -> cool-stuff
        What?? -> what
    """
    # Convert everything to lowercase
    name = str(name).lower()
    # Replace period delimiters with single space
    name = re.sub(r"\.+", " ", name)
    # Strip non-word characters
    name = re.sub(r"[^\w\s]", "", name)
    # Strip leading/trailing whitespace
    name = name.strip()
    # Convert whitespace runs to single dash
    name = re.sub(r"\s+", "-", name)

    return name


class ValidationError(ValueError):
    """Validation error from malformed database or locator entry."""


class LocatorsDatabase:
    """Container for storing locator information,
    and serializing/deserializing database file.
    """

    def __init__(self, path: Optional[Union[str, io.IOBase]] = None):
        self.logger = logging.getLogger(__name__)
        self.path = path or self.default_path
        self.locators = {}
        self._error = None
        self._invalid = {}

    @classmethod
    def load_by_name(cls, name, path: Optional[str] = None) -> Locator:
        """Load locator entry from database with given name."""
        database = cls(path)
        database.load()

        if database.error:
            error_msg, error_args = database.error
            raise ValueError(error_msg % error_args)

        return database.resolve(name)

    @property
    def default_path(self) -> str:
        """Return default path for locators database file."""
        dirname = os.getenv("ROBOT_ROOT", "")
        filename = os.getenv("RPA_LOCATORS_DATABASE", "").strip() or "locators.json"
        return os.path.join(dirname, filename)

    @property
    def parent(self):
        """Return parent directory for database."""
        return (
            Path(self.path).parent
            if not isinstance(self.path, io.IOBase)
            else Path(".")
        )

    @property
    def error(self):
        return self._error

    def resolve(self, name):
        """Fetch locator form database, and fix relative paths."""
        if name not in self.locators:
            raise ValueError(f"No locator with name: {name}")

        locator = self.locators[name]

        def to_absolute(field_name):
            value = getattr(locator, field_name)
            setattr(locator, field_name, self._resolve_path(value))

        if isinstance(locator, BrowserLocator):
            to_absolute("screenshot")
        elif isinstance(locator, ImageLocator):
            to_absolute("path")
            to_absolute("source")

        return locator

    def _resolve_path(self, value):
        if value is not None and not Path(value).is_absolute():
            return str(Path(self.parent) / value)
        return value

    def set_error(self, msg, *args):
        """Log an error message. Ensures the same message
        is not repeated multiple times.
        """
        message = (msg, args)
        if message != self._error:
            self.logger.error(msg, *args)
            self._error = message

    def reset_error(self):
        """Clear error state."""
        self._error = None

    def load(self):
        """Deserialize database from file."""
        try:
            with open_stream(self.path, "r") as fd:
                data = json.load(fd)
                locators, invalid = self._load(data)

            self.locators = locators
            self._invalid = invalid
            self.reset_error()
        except FileNotFoundError:
            self.locators = {}
            self._invalid = {}
            self.reset_error()
        except Exception as err:  # pylint: disable=broad-except
            self.locators = {}
            self._invalid = {}
            self.set_error("Could not read database: %s", str(err))

    def save(self):
        """Serialize database into file."""
        data = {}

        for name, locator in self.locators.items():
            data[name] = locator.to_dict()

        for name, fields in self._invalid.items():
            if name not in data:
                data[name] = fields

        with open_stream(self.path, "w") as fd:
            json.dump(data, fd, sort_keys=True, indent=4)

    def _load(self, data):
        """Load database content as Locator objects."""
        locators = {}
        invalid = {}

        # Migrate from old list-based file format
        if isinstance(data, list):
            data = {fields["name"]: fields for fields in data if "name" in fields}

        sanitized = []
        for name, fields in data.items():
            try:
                sname = sanitize_name(name)
                if sname in sanitized:
                    raise ValueError(f"Duplicate sanitized name: {name} / {sname}")
                sanitized.append(sname)

                locators[name] = Locator.from_dict(fields)
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.warning('Failed to parse locator "%s": %s', name, exc)
                invalid[name] = fields

        # Migrate base64 screenshots to files
        for name, locator in locators.items():
            self._convert_screenshot(name, locator)

        return locators, invalid

    def _convert_screenshot(self, name, locator):
        """Migrate base64 screenshot to file."""
        if not isinstance(locator, BrowserLocator) or not locator.screenshot:
            return

        try:
            path = self._resolve_path(locator.screenshot)
            if Path(path).is_file():
                return
        except Exception:  # pylint: disable=broad-except
            pass

        try:
            content = base64.b64decode(locator.screenshot)
        except Exception:  # pylint: disable=broad-except
            self.logger.info("Invalid screenshot: %s", locator.screenshot)
            return

        images = self.parent / ".images"
        path = images / "{}-{}.png".format(sanitize_name(name), "screenshot")

        os.makedirs(images, exist_ok=True)
        with open(path, "wb") as fd:
            fd.write(content)

        locator.screenshot = str(path.relative_to(self.parent))



from pathlib import Path
from typing import Dict, List, Optional, Union

from RPA.core.windows import WindowsElements
from RPA.core.windows.helpers import IS_WINDOWS
from RPA.core.windows.locators import MatchObject, WindowsElement
from RPA.core.windows.window import WindowMethods

if IS_WINDOWS:
    import uiautomation as auto
    from uiautomation import Control  # pylint: disable=unused-import


RecordElement = Dict[str, Optional[Union[float, str, "Control", List[str]]]]

NOT_AVAILABLE = "N/A"


class ElementInspector:
    """Element locator inspector"""

    MATCH_PRIORITY = ["automation_id", "name"]  # element matching priority

    def __init__(self):
        # Lazily loaded with verbose mode on for printing the tree and returning the
        #  structure.
        self._windows_elements: Optional[WindowsElements] = None

    @property
    def windows_elements(self) -> WindowsElements:
        """The minimal core flavor of the Windows library."""
        if not self._windows_elements:
            self._windows_elements = WindowsElements()
        return self._windows_elements

    def inspect_element(
        self,
        recording: List[RecordElement],
        verbose: bool = False,
    ) -> None:
        """Inspect Windows element under mouse pointer.

        :param recording: Store the dict records under this list.
        :param verbose: Show exhaustive locators if `True`, otherwise just simple ones.
            Switching this on will make recording slower as it is refreshing the
            element tree with each click in order to provide their path strategy as
            well.
        """
        # TODO(cmin764): Support Python syntax as well. (currently just RF keywords)
        with auto.UIAutomationInitializerInThread(debug=False):
            control = auto.ControlFromCursor()
            parent_control = control.GetParentControl()
            exec_path = ""
            try:
                top_level_control = control.GetTopLevelControl()
            except AttributeError:
                top_level_control = None
                top_level_handle = NOT_AVAILABLE
            else:
                top_level_handle = top_level_control.NativeWindowHandle
                try:
                    exec_path = WindowMethods.get_fullpath(top_level_control.ProcessId)
                except Exception:  # pylint: disable=broad-except
                    pass

            top_properties = self._get_element_key_properties(
                top_level_control, verbose=verbose
            )
            parent_properties = self._get_element_key_properties(
                parent_control, verbose=verbose
            )
            child_properties = self._get_element_key_properties(
                control, top_level_control=top_level_control, verbose=verbose
            )

            top_locator = " and ".join(top_properties) or NOT_AVAILABLE
            parent_locator = " and ".join(parent_properties) or NOT_AVAILABLE
            child_locator = " and ".join(child_properties) or NOT_AVAILABLE
            control_locator = child_locator
            if parent_locator != NOT_AVAILABLE and not (
                "name:" in child_locator or "id:" in child_locator
            ):
                control_locator = f"{parent_locator} > {child_locator}"

            recording.append(
                {
                    "type": "locator",
                    "exec_path": exec_path,
                    "exec": Path(exec_path).name,
                    "top": top_locator,
                    "top_handle": top_level_handle,
                    "x": top_level_control,
                    "locator": control_locator,
                    "top_props": top_properties,
                    "parent_props": parent_properties,
                    "props": child_properties,
                    "name": parent_control.Name if parent_control else None,
                    "control": parent_control,
                }
            )

    @staticmethod
    def _get_locators(*, regex_limit: int, **kwargs) -> List[str]:
        locators = []

        name = kwargs["name"]
        if name:
            name_property = "name:"
            if len(name) > regex_limit:
                name_property = "regex:"
                name = name[:regex_limit].strip()
            if " " in name:
                q = MatchObject.QUOTE
                name = f"{q}{name}{q}"
            locators.append(f"{name_property}{name}")

        automation_id, control_type, class_name = (
            kwargs["automation_id"],
            kwargs["control_type"],
            kwargs["class_name"],
        )
        # NOTE(cmin764): Sometimes, the automation ID is a randomly generated number,
        #  different with each run. (therefore you can't rely on it in the locator)
        if automation_id and not str(automation_id).isnumeric():
            locators.append(f"id:{automation_id}")
        if control_type:
            locators.append(f"type:{control_type}")
        if class_name:
            locators.append(f"class:{class_name}")

        return locators

    @classmethod
    def _filter_elements(
        cls,
        elements: List[WindowsElement],
        *,
        control_type: str,
        class_name: str,
        **kwargs,
    ) -> Dict[str, List[WindowsElement]]:
        candidates: Dict[str, List[WindowsElement]] = {}

        for element in elements:
            not_good = (
                control_type
                and control_type != element.control_type
                or class_name
                and class_name != element.class_name
            )
            if not_good:
                continue

            for attr in cls.MATCH_PRIORITY:
                value = kwargs[attr]
                if value and value == getattr(element, attr):
                    candidates.setdefault(attr, []).append(element)

        return candidates

    def _match_element_for_path(
        self, control: "Control", top_level_control: "Control", **kwargs
    ) -> Optional[str]:
        # Compute how deep the search should go.
        at_level = 0
        cursor = control
        while not cursor.IsTopLevel():
            cursor = cursor.GetParentControl()
            at_level += 1
        if at_level == 0:
            return None  # the clicked element is the root of the tree (main window)

        # Obtain a new element tree structure during every click, as the tree changes
        #  (expands/shrinks/rotates) with element actions producing UI display changes.
        top_level_element = WindowsElement(top_level_control, None)
        structure = self.windows_elements.print_tree(
            top_level_element,
            return_structure=True,
            log_as_warnings=None,
            max_depth=at_level,
        )
        elements_dict = self._filter_elements(structure[at_level], **kwargs)

        for prio in self.MATCH_PRIORITY:
            elements = elements_dict.get(prio, [])
            for candidate in elements:
                maybe_path = candidate.locator.rsplit(MatchObject.TREE_SEP, 1)[-1]
                if "path:" in maybe_path:
                    return maybe_path

        return None

    def _get_element_key_properties(
        self,
        control: Optional["Control"],
        *,
        top_level_control: Optional["Control"] = None,
        verbose: bool,
        regex_limit: int = 300,
    ) -> List[str]:
        if not control:
            print("Got null control!")
            return []

        props = {
            "name": control.Name,
            "automation_id": control.AutomationId,
            "control_type": control.ControlTypeName,
            "class_name": control.ClassName,
        }
        locators = self._get_locators(regex_limit=regex_limit, **props)

        # Add the `path:` strategy as well with verbose recordings. (useful when you
        #  can't rely on Automation IDs nor names)
        if verbose and top_level_control:
            path = self._match_element_for_path(
                control,
                top_level_control,
                **props,
            )
            if path:
                locators.append(path)

        if locators:
            if not verbose:
                locators = locators[:1]
            return locators

        print("Was unable to construct locator for the control!")
        return []
